package edu.washington.cs.atomicdebug.examples;

import edu.washington.cs.atomicdebug.AtomicDebug;

/**
 * Bank account transfers are a nice example of why we want composable atomic
 * transactions. If transfer() uses an atomic block, and also calls deposit()
 * and withdraw() which also use atomic blocks, no problem - the atomic blocks
 * compose seamlessly. The version with locks may result in deadlock.
 * 
 * This "solution" has some thread sleeps and atomic blocks added to demonstrate
 * and fix race conditions.
 * 
 * @author Laura Effinger-Dean
 */
public class BankAccountSolved {
	private int balance;

	/**
	 * Initialize the bank account with the given balance.
	 * 
	 * @param balance
	 *            Initial balance.
	 */
	public BankAccountSolved(int balance) {
		this.balance = balance;
	}

	public int balance() {
		AtomicDebug.enterAtomic();
		int result = balance;
		AtomicDebug.exitAtomic();
		return result;
	}

	public void withdraw(int amount) {
		try {
			AtomicDebug.enterAtomic();
			int temp = balance - amount;
			Thread.sleep(10);
			balance = temp;
			AtomicDebug.exitAtomic();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	public void deposit(int amount) {
		try {
			AtomicDebug.enterAtomic();
			int temp = balance + amount;
			Thread.sleep(10);
			balance = temp;
			AtomicDebug.exitAtomic();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	public void transfer(BankAccountSolved toAccount, int amount) {
		try {
			AtomicDebug.enterAtomic();
			this.withdraw(amount);
			Thread.sleep(1);
			toAccount.deposit(amount);
			AtomicDebug.exitAtomic();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
